'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _slicedToArray = function () { function sliceIterator(arr, i) { var _arr = []; var _n = true; var _d = false; var _e = undefined; try { for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) { _arr.push(_s.value); if (i && _arr.length === i) break; } } catch (err) { _d = true; _e = err; } finally { try { if (!_n && _i["return"]) _i["return"](); } finally { if (_d) throw _e; } } return _arr; } return function (arr, i) { if (Array.isArray(arr)) { return arr; } else if (Symbol.iterator in Object(arr)) { return sliceIterator(arr, i); } else { throw new TypeError("Invalid attempt to destructure non-iterable instance"); } }; }();

exports.run = run;
exports.caseOf = caseOf;

var _symbolObservable = require('symbol-observable');

var _symbolObservable2 = _interopRequireDefault(_symbolObservable);

var _Observable = require('rxjs/Observable');

var _Subject = require('rxjs/Subject');

require('rxjs/add/operator/startWith');

require('rxjs/add/operator/scan');

require('rxjs/add/operator/do');

require('rxjs/add/operator/map');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function isEvent(e) {
  return e && e.preventDefault && e.stopPropagation;
}

var hasRaf = typeof window !== 'undefined' && window.requestAnimationFrame;

function run(program, render) {
  var opt = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  var _program$main = program.main,
      init = _program$main.init,
      view = _program$main.view,
      update = _program$main.update,
      subscriptions = _program$main.subscriptions;
  var onUpdate = opt.onUpdate;

  var msg$ = new _Subject.Subject();
  var initModel = init,
      sideEffect = void 0;

  // Use requestAnimationFrame for better render performance
  var rafRender = hasRaf ? function (vdom) {
    return requestAnimationFrame(function () {
      return render(vdom);
    });
  } : render;

  var platform = _defineProperty({
    cmd: msg$.next.bind(msg$),
    subscribe: msg$.subscribe.bind(msg$)
  }, _symbolObservable2.default, function () {
    return platform;
  });

  // alias
  platform.sub = platform.subscribe;

  if (typeof initModel === 'undefined') {
    throw new Error('Initial model is undefined. Pass null to explicitly set `no value`.');
  }

  if (Array.isArray(initModel) && typeof initModel[1] === 'function') {
    var _initModel = initModel;

    var _initModel2 = _slicedToArray(_initModel, 2);

    initModel = _initModel2[0];
    sideEffect = _initModel2[1];
  }

  var model$ = msg$.startWith(initModel).scan(function (model, msg) {
    var newModel = update(msg, model);

    if (typeof newModel === 'undefined') {
      throw new Error('Unhandled msg ' + msg.constructor.name + ' ' + JSON.stringify(msg));
    }

    if (Array.isArray(newModel) && typeof newModel[1] === 'function') {
      var _newModel = newModel;

      var _newModel2 = _slicedToArray(_newModel, 2);

      newModel = _newModel2[0];
      sideEffect = _newModel2[1];
    }

    onUpdate && onUpdate(msg, model, newModel);
    return newModel;
  })._do(function () {
    sideEffect && sideEffect(platform);
    sideEffect = null;
  });

  var view$ = view(dispatch, model$);
  if (typeof view$ === 'function') {
    view$ = model$.map(view$);
  }

  view$.subscribe(rafRender);

  subscriptions && subscriptions(platform, initModel);

  function dispatch(msg) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    if (typeof msg === 'function') {
      return function eventHandler(e) {
        if (isEvent(e)) {
          var type = e.type,
              target = e.target;

          if (type === 'change' && target.tagName === 'INPUT' && (target.type === 'radio' || target.type === 'checkbox')) {
            msg$.next(new (Function.prototype.bind.apply(msg, [null].concat(args, [target.checked])))());
          } else {
            msg$.next(new (Function.prototype.bind.apply(msg, [null].concat(args, [target.value])))());
          }
        } else {
          msg$.next(new (Function.prototype.bind.apply(msg, [null].concat(args)))());
        }
      };
    }

    msg$.next(msg);
  }

  return {
    platform: platform,
    terminate: msg$.complete.bind(msg$)
  };
}

function caseOf(msg) {
  for (var i = 0; i < (arguments.length <= 1 ? 0 : arguments.length - 1); i += 2) {
    var Msg = arguments.length <= i + 1 ? undefined : arguments[i + 1];
    var update = arguments.length <= i + 1 + 1 ? undefined : arguments[i + 1 + 1];
    if (msg.constructor === Msg) {
      return update(msg);
    }
  }
}